#!/usr/bin/env python
#==============================================================================
# Copyright and license info is available in the LICENSE file included with
# the Server Deployment Package (SDP), and also available online:
# https://swarm.workshop.perforce.com/projects/perforce-software-sdp/view/main/LICENSE
#------------------------------------------------------------------------------

#
# This trigger script ensures that 'Map:' field on a depot spec is not the 
# unmodified server default, which stores values under the P4ROOT folder.
# The value is set to the standard SDP location for depot storage, but only
# if the current value is the unmodified server default.
#
# Enable this in the triggers table.  On a Linux server, the entry looks
# like this:
#
#   DepotSpecMapField form-out depot "/p4/common/bin/triggers/SetDefaultDepotSpecMapField.py %formfile% 1"
#
# The second paramter, a 1 (one) in the example above, is the instance number.
# If there are multiple Perforce server instances, this trigger should be
# installed in each instance, with the appropriate instance number.
#
# On a Windows server, the triggers table requires an extra argument.  It looks
# like this:
#
#   DepotSpecMapField form-out depot "python <depotdrive>\p4\common\bin\triggers\SetDefaultDepotSpecMapField.py %formfile% 1 <depotdrive>"
#
# where the second parameter is the instance number, and <depotdrive> is the
# drive letter for your depot volume, such as F:
#
# See 'p4 help triggers' for more info.

from __future__ import print_function

import os
import sys
import re
import random
import platform
import shutil

def cleanexit():
  input.close()
  output.close()
  os.remove(tempfile)
  sys.exit(0)

def argserror():
  print("Invalid number of parameters supplied, see comments for instructions.")
  sys.exit(1)

###############################################################################
# main
if __name__ == "__main__":
  OS = platform.system()
  tempfile = str(random.random())

  if OS == "Windows":
    if len(sys.argv) < 4:
      argserror()
    else:
      DepotDrive = sys.argv[3]
  else:
    if len(sys.argv) < 3:
      argserror()
    else:
      DepotDrive = ""

  DepotSpecFile = sys.argv[1]
  Instance = sys.argv[2]
  mapline = "^Map:\t%s/p4/%s" % (DepotDrive, Instance)
  maptext = r"Map:      %s/p4/%s/depots/" % (DepotDrive, Instance)

  input = open(DepotSpecFile, "r")
  output = open(tempfile, "w")

  for line in input.readlines():
    if re.search("^Depot:\t", line):
      DepotName = re.sub("^Depot:\t", "", line)
      DepotName = DepotName.strip()
    if re.search("^Type:\tremote", line):
      cleanexit()
    if re.search(mapline, line, re.IGNORECASE):
      cleanexit()
    if re.search("^Map:\t", line):
      if not re.search("^Map:\t%s\/\.\.\." % DepotName, line):
        cleanexit()
      line = maptext + DepotName + "/..."
      output.write(line)
    else:
      output.write(line)

  input.close()
  output.close()

  os.remove(DepotSpecFile)
  shutil.copyfile(tempfile, DepotSpecFile)
  os.remove(tempfile)

  sys.exit(0)

