#!/bin/bash
#==============================================================================
# Copyright and license info is available in the LICENSE file included with
# the Server Deployment Package (SDP), and also available online:
# https://swarm.workshop.perforce.com/projects/perforce-software-sdp/view/main/LICENSE
#------------------------------------------------------------------------------
export SDP_INSTANCE=${SDP_INSTANCE:-Unset} 
export SDP_INSTANCE=${1:-$SDP_INSTANCE} 
if [[ $SDP_INSTANCE == Unset ]]; then 
   echo -e "\nError: Instance parameter not supplied.\n" 
   echo "You must supply the Perforce instance as a parameter to this script." 
   exit 1 
fi 

# Load SDP controlled shell environment.
# shellcheck disable=SC1091
source /p4/common/bin/p4_vars "$SDP_INSTANCE" ||\
   { echo -e "\nError: Failed to load SDP environment.\n"; exit 1; }

# shellcheck disable=SC1090
source "$P4CBIN/backup_functions.sh" ||\
   { echo -e "\nError: Failed to load SDP backup_functions.sh lib.\n"; exit 1; }

cd "$P4CBIN" || { echo -e "\nError: Failed to cd to $P4CBIN.\n"; exit 1; }

######### Start of Script ##########

check_vars
set_vars

# shellcheck disable=SC2155
export LOGFILE="${LOGS:-/tmp}/upgrade.$P4SERVER.$(date +'%Y%m%d-%H%M').log"

echo "Processing. This may take a few minutes depending on time to rotate the active journal and replay to the offline_db."
echo "Log file is: $LOGFILE"

check_dirs
check_uid

[[ -f "$P4CBIN/p4" ]] || { echo "No p4 in $P4CBIN" ; exit 1 ;}
[[ -f "$P4CBIN/p4d" ]] || { echo "No p4d in $P4CBIN" ; exit 1 ;}

chmod 755 "$P4CBIN/p4"
chmod 700 "$P4CBIN/p4d"
[[ -f "$P4CBIN/p4broker" ]] && chmod 700 "$P4CBIN/p4broker"
[[ -f "$P4CBIN/p4p" ]] && chmod 700 "$P4CBIN/p4p"

#shellcheck disable=SC2016
P4RELNUM=$(./p4 -V | grep -i Rev. | $AWK -F / '{print $3}')
#shellcheck disable=SC2016
P4DRELNUM=$(./p4d -V | grep -i Rev. | $AWK -F / '{print $3}')
#shellcheck disable=SC2016
P4BLDNUM=$(./p4 -V | grep -i Rev. | $AWK -F / '{print $4}' | awk '{print $1}')
#shellcheck disable=SC2016
P4DBLDNUM=$(./p4d -V | grep -i Rev. | $AWK -F / '{print $4}' | awk '{print $1}')
#shellcheck disable=SC2016
CURRENT_RELNUM=$("./p4d_${SDP_INSTANCE}_bin" -V | grep -i Rev. | $AWK -F / '{print $3}')

log "Start $P4SERVER Upgrade on ${HOSTNAME%%.*} at $(date)."
log "Current server release: $CURRENT_RELNUM"
log "New server release: $P4DRELNUM.$P4DBLDNUM"

[[ -f "p4_$P4RELNUM.$P4BLDNUM" ]] || cp p4 "p4_$P4RELNUM.$P4BLDNUM"
[[ -f "p4d_$P4DRELNUM.$P4DBLDNUM" ]] || cp p4d "p4d_$P4DRELNUM.$P4DBLDNUM"
[[ -f "p4_${P4RELNUM}_bin" ]] && unlink "p4_${P4RELNUM}_bin"
ln -s "p4_$P4RELNUM.$P4BLDNUM" "p4_${P4RELNUM}_bin" >> "$LOGFILE" 2>&1
[[ -f "p4d_${P4DRELNUM}_bin" ]] && unlink "p4d_${P4DRELNUM}_bin"
ln -s "p4d_$P4DRELNUM.$P4DBLDNUM" "p4d_${P4DRELNUM}_bin" >> "$LOGFILE" 2>&1
[[ -f p4_bin ]] && unlink p4_bin
ln -s "p4_${P4RELNUM}_bin" p4_bin >> "$LOGFILE" 2>&1

if [[ -L "p4broker_${SDP_INSTANCE}_bin" ]]; then
   # shellcheck disable=SC2154
   if [[ -x "$P4BrokerInitScript" && \
         -f "/p4/common/config/p4_${SDP_INSTANCE}.broker.cfg" ]]; then
      stop_p4broker
   fi

   # shellcheck disable=SC2016
   P4BRELNUM=$(./p4broker -V | grep -i Rev. | $AWK -F / '{print $3}')
   # shellcheck disable=SC2016
   P4BBLDNUM=$(./p4broker -V | grep -i Rev. | $AWK -F / '{print $4}' | $AWK '{print $1}')

   [[ -f "p4broker_$P4BRELNUM.$P4BBLDNUM" ]] || cp p4broker "p4broker_$P4BRELNUM.$P4BBLDNUM"
   [[ -L "p4broker_${P4BRELNUM}_bin" ]] && unlink "p4broker_${P4BRELNUM}_bin"
   ln -s "p4broker_$P4BRELNUM.$P4BBLDNUM" "p4broker_${P4BRELNUM}_bin"
   [[ -L "p4broker_${SDP_INSTANCE}_bin" ]] && unlink "p4broker_${SDP_INSTANCE}_bin"
   ln -s "p4broker_${P4BRELNUM}_bin" "p4broker_${SDP_INSTANCE}_bin"
   if [[ -x "$P4BrokerInitScript" && \
         -f "/p4/common/config/p4_${SDP_INSTANCE}.broker.cfg" ]]; then
      start_p4broker
   fi
fi

if [[ -L "p4p_${SDP_INSTANCE}_bin" ]]; then
   # shellcheck disable=SC2154
   if [[ -x "$P4ProxyInitScript" ]]; then
      stop_p4p
   fi

   # shellcheck disable=SC2016
   P4PRELNUM=$(./p4p -V | grep -i Rev. | $AWK -F / '{print $3}')
   # shellcheck disable=SC2016
   P4PBLDNUM=$(./p4p -V | grep -i Rev. | $AWK -F / '{print $4}' | $AWK '{print $1}')
   [[ -f "p4p_$P4PRELNUM.$P4PBLDNUM" ]] || cp p4p "p4p_$P4PRELNUM.$P4PBLDNUM"
   [[ -L "p4p_${P4PRELNUM}_bin" ]] && unlink "p4p_${P4PRELNUM}_bin"
   ln -s "p4p_$P4PRELNUM.$P4PBLDNUM" "p4p_${P4PRELNUM}_bin"
   [[ -L "p4p_${SDP_INSTANCE}_bin" ]] && unlink "p4p_${SDP_INSTANCE}_bin"
   ln -s "p4p_${P4PRELNUM}_bin" "p4p_${SDP_INSTANCE}_bin"
   if [[ -x "$P4ProxyInitScript" ]]; then
      start_p4p
   fi
fi

"$P4CBIN/p4login" -v >> "$LOGFILE" 2>&1

if [[ "$P4REPLICA" == "FALSE" ]] || [[ $EDGESERVER -eq 1 ]]; then
   get_journalnum
fi

stop_p4d

if [[ "$CURRENT_RELNUM" != "$P4DRELNUM" ]]; then
   # Only rotate the journal on the master. 
   if [[ "$P4REPLICA" == "FALSE" ]] && [[ $EDGESERVER -eq 0 ]]; then
      p4d_truncate_journal
      get_offline_journal_num
      replay_journals_to_offline_db
   fi
   unlink "p4d_${SDP_INSTANCE}_bin" >> "$LOGFILE" 2>&1
   ln -s "p4d_${P4DRELNUM}_bin" "p4d_${SDP_INSTANCE}_bin" >> "$LOGFILE" 2>&1

   log "Upgrading databases in $P4ROOT."
   # For versions >= 19.1 we should journal the upgrade and not upgrade offline
   # In such cases we let offline_db be upgraded by applying rotate journals in the
   # normal way.
   UPGRADE_JNL="$P4JOURNAL"
   # shellcheck disable=SC2072
   if [[ "${P4DRELNUM}.${P4DBLDNUM}" < "18.2" ]]; then
      UPGRADE_JNL=off
   fi

   "$P4DBIN" -r "$P4ROOT" -J "$UPGRADE_JNL" -xu >> "$LOGFILE" 2>&1 ||\
      die "Failed to upgrade root database."

   # Upgrade the offline database on the master and on edge servers.
   if [[ "$P4REPLICA" == "FALSE" ]] || [[ "$EDGESERVER" -eq 1 ]]; then
      if [[ "$UPGRADE_JNL" == "off" ]]; then
         log "Upgrading databases in $OFFLINE_DB as this is a pre 18.2 server upgrade"

         # The value of '-t localhost:0000' is necessary for syntax reasons;
         # it is ignored. This allows upgade of offline databases without needing
         # a license file; there isn't one in offline_db, only P4ROOT.  A valid
         # license file is required to actually start the server process and have
         # it listen on a port.

         "$P4DBIN" -r "$OFFLINE_DB" -J "$UPGRADE_JNL" -xu -t localhost:0000 >> "$LOGFILE" 2>&1 || die "Failed to upgrade offline_db"
      else
         if [[ "$P4REPLICA" == "FALSE" ]] && [[ "$EDGESERVER" -eq 0 ]]; then
            log "Upgrade offline through rotating latest journal on master/commit server"
            get_journalnum
            p4d_truncate_journal
            get_offline_journal_num
            replay_journals_to_offline_db
         fi
      fi
   fi
fi

start_p4d

if [[ "$P4REPLICA" == "FALSE" && "$EDGESERVER" -eq 0 ]]; then
   log "Upgrading protections table comment format."
   "$P4BIN" protect --convert-p4admin-comments -o | "$P4BIN" -s protect -i >> "$LOGFILE" 2>&1
fi

log "Finished $P4SERVER Upgrade"

cat "$LOGFILE"
